@@ -1,6 +1,7 @@
 use crate::gen::SchemaGenerator;
 use crate::schema::*;
 use crate::JsonSchema;
+use std::borrow::Cow;
 use url::Url;
 
 impl JsonSchema for Url {
@@ -10,6 +11,10 @@ impl JsonSchema for Url {
         "Url".to_owned()
     }
 
+    fn schema_id() -> Cow<'static, str> {
+        Cow::Borrowed("url::Url")
+    }
+
     fn json_schema(_: &mut SchemaGenerator) -> Schema {
         SchemaObject {
             instance_type: Some(InstanceType::String.into()),
